@@ -1,19 +1,30 @@
 /*
-  copyright   : GPL
+  Copyright Ⓒ 1997, 1998, 1999, 2000, 2001  joost witteveen
+  Copyright Ⓒ 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009  Clint Adams
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
   title       : fakeroot
-  description : create a "fake" root shell, by wrapping 
+  description : create a "fake" root shell, by wrapping
                 functions like chown, stat, etc. Useful for debian
                 packaging mechanism
-  author      : joost witteveen, joostje@debian.org
 */
 
 /*
-  upon startup, the fakeroot script (/usr/bin/fakeroot) 
+  upon startup, the fakeroot script (/usr/bin/fakeroot)
   forks faked (this program), and the shell or user program that
   will run with the libtricks.so.0.0 wrapper.
-  
+
   These tree running programs have the following tasks:
-    
+
     fakeroot script
        starts the other two processes, waits for the user process to
        die, and then send a SIGTERM signal to faked, causing
@@ -22,13 +33,13 @@
     faked
        the ``main'' daemon, creates ipc message queues, and later
        receives ipc messages from the user program, maintains
-       fake inode<->ownership database (actually just a 
+       fake inode<->ownership database (actually just a
        lot of struct stat entries). Will clear ipc message ques
        upon receipt of a SIGTERM. Will show debug output upon
        receipt of a SIGUSR1 (if started with -d debug option)
 
     user program
-       Any shell or other programme, run with 
+       Any shell or other programme, run with
        LD_PRELOAD=libtricks.so.0.0, and FAKEROOT_DBKEY=ipc-key,
        thus the executed commands will communicate with
        faked. libtricks will wrap all file ownership etc modification
@@ -40,11 +51,11 @@
      getuid(), geteuid(), getgid(), getegid(),
      mknod()
      chown(), fchown() lchown()
-     chmod(), fchmod() 
+     chmod(), fchmod()
      mkdir(),
      lstat(), fstat(), stat() (actually, __xlstat, ...)
      unlink(), remove(), rmdir(), rename()
-    
+
   comments:
     I need to wrap unlink because of the following:
         install -o admin foo bar
@@ -56,18 +67,26 @@
     Same goes for all other ways to remove inodes form the filesystem,
     like rename(existing_file, any_file).
 
-    The communication between client (user progamme) and faked happens 
-    with inode/dev information, not filenames. This is 
+    The communication between client (user progamme) and faked happens
+    with inode/dev information, not filenames. This is
     needed, as the client is the only one who knows what cwd is,
     so it's much easier to stat in the client. Otherwise, the daemon
     needs to keep a list of client pids vs cwd, and I'd have to wrap
     fork e.d., as they inherit their parent's cwd. Very compilcated.
-    
+
     */
-/* ipc documentation bugs: msgsnd(2): MSGMAX=4056, not 4080 
-   (def in ./linux/msg.h, couldn't find other def in /usr/include/ 
+/* ipc documentation bugs: msgsnd(2): MSGMAX=4056, not 4080
+   (def in ./linux/msg.h, couldn't find other def in /usr/include/
    */
 
+#ifdef __APPLE__
+/*
+   In this file, we want 'struct stat' to have a 32-bit 'ino_t'.
+   We use 'struct stat64' when we need a 64-bit 'ino_t'.
+*/
+#define _DARWIN_NO_64_BIT_INODE
+#endif
+
 #include "config.h"
 #include "communicate.h"
 #ifndef FAKEROOT_FAKENET
@@ -85,6 +104,9 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <sys/types.h>
+#ifdef HAVE_SYS_XATTR_H
+#include <sys/xattr.h>
+#endif
 #include <fcntl.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -131,6 +153,10 @@ void process_chmod(struct fake_msg *buf);
 void process_mknod(struct fake_msg *buf);
 void process_stat(struct fake_msg *buf);
 void process_unlink(struct fake_msg *buf);
+void process_listxattr(struct fake_msg *buf);
+void process_setxattr(struct fake_msg *buf);
+void process_getxattr(struct fake_msg *buf);
+void process_removexattr(struct fake_msg *buf);
 
 #ifdef FAKEROOT_FAKENET
 static int get_fakem(struct fake_msg *buf);
@@ -143,6 +169,12 @@ process_func func_arr[]={process_chown,
 			 process_mknod,
 			 process_stat,
 			 process_unlink,
+			 NULL, /* debugdata */
+			 NULL, /* reqoptions */
+			 process_listxattr,
+			 process_getxattr,
+			 process_setxattr,
+			 process_removexattr,
 			 };
 
 unsigned int highest_funcid = sizeof(func_arr)/sizeof(func_arr[0]);
@@ -171,13 +203,91 @@ static void fail(const char *msg)
 }
 #endif
 
+struct xattr_node_s;
+typedef struct xattr_node_s {
+  struct xattr_node_s *next;
+  char                *key;
+  char                *value;
+  size_t              value_size;
+} xattr_node_t;
+
 struct data_node_s;
 typedef struct data_node_s {
   struct data_node_s *next;
   struct fakestat     buf;
   uint32_t            remote;
+  xattr_node_t       *xattr;
 } data_node_t;
 
+static xattr_node_t *xattr_find(xattr_node_t *node, char *key)
+{
+  while (node) {
+    if (node->key && (!strcmp(node->key, key)))
+      break;
+    node = node->next;
+  }
+  return node;
+}
+
+static int xattr_erase(xattr_node_t **head, char *key)
+{
+  xattr_node_t *cur_node, *prev_node = NULL;
+
+  for (cur_node = *head; cur_node; prev_node = cur_node, cur_node = cur_node->next) {
+    if (cur_node->key && (!strcmp(cur_node->key, key))) {
+      if (prev_node == NULL) {
+        *head = cur_node->next;
+      } else {
+        prev_node->next = cur_node->next;
+      }
+      free(cur_node->key);
+      if (cur_node->value)
+        free(cur_node->value);
+      free(cur_node);
+      return 1;
+    }
+  }
+  return 0;
+}
+
+static int xattr_clear(xattr_node_t **head)
+{
+  xattr_node_t *cur_node, *next_node;
+  cur_node = *head;
+  *head = NULL;
+  while (cur_node) {
+    next_node = cur_node->next;
+    if (cur_node->key)
+      free(cur_node->key);
+    if (cur_node->value)
+      free(cur_node->value);
+    free(cur_node);
+    cur_node = next_node;
+  }
+  return 0;
+}
+
+static xattr_node_t *xattr_insert(xattr_node_t **head)
+{
+  xattr_node_t *new_node;
+  new_node = calloc(1, sizeof(xattr_node_t));
+  new_node->next = *head;
+  *head = new_node;
+  return new_node;
+}
+
+static void xattr_fill(xattr_node_t *node, char *key, char *value, size_t value_size)
+{
+  if (node->key)
+    free(node->key);
+  node->key = strdup(key);
+  if (node->value)
+    free(node->value);
+  node->value = malloc(value_size);
+  memcpy(node->value, value, value_size);
+  node->value_size = value_size;
+}
+
 #define data_node_get(n)   ((struct fakestat *) &(n)->buf)
 
 #define HASH_TABLE_SIZE 10009
@@ -213,7 +323,7 @@ static void data_insert(const struct fakestat *buf,
 			const uint32_t remote)
 {
   data_node_t *n, *last = NULL;
-  
+
   for (n = data_hash_table[data_hash_val(buf)]; n; last = n, n = n->next)
     if (fakestat_equal(&n->buf, buf) && n->remote == remote)
       break;
@@ -228,6 +338,7 @@ static void data_insert(const struct fakestat *buf,
   }
 
   memcpy(&n->buf, buf, sizeof (struct fakestat));
+  n->xattr = NULL;
   n->remote = (uint32_t) remote;
 }
 
@@ -247,6 +358,7 @@ static data_node_t *data_erase(data_node_t *pos)
   else
     prev->next = next;
 
+  xattr_clear(&n->xattr);
   free(n);
 
   return next;
@@ -342,6 +454,9 @@ static void faked_send_fakem(const struct fake_msg *buf)
   fm.st.mode = htonl(buf->st.mode);
   fm.st.nlink = htonl(buf->st.nlink);
   fm.remote = htonl(buf->remote);
+  fm.xattr.buffersize = htonl(buf->xattr.buffersize);
+  fm.xattr.flags_rc = htonl(buf->xattr.flags_rc);
+  memcpy(fm.xattr.buf, buf->xattr.buf, MAX_IPC_BUFFER_SIZE);
 
   while (1) {
     ssize_t len;
@@ -580,7 +695,7 @@ void debug_stat(const struct fakestat *st){
 void insert_or_overwrite(struct fakestat *st,
 			 const uint32_t remote){
   data_node_t *i;
-  
+
   i = data_find(st, remote);
   if (i == data_end()) {
     if(debug){
@@ -604,7 +719,7 @@ void process_chown(struct fake_msg *buf){
   struct fakestat *stptr;
   struct fakestat st;
   data_node_t *i;
-  
+
   if(debug){
     fprintf(stderr,"FAKEROOT: chown ");
     debug_stat(&buf->st);
@@ -612,11 +727,11 @@ void process_chown(struct fake_msg *buf){
   i = data_find(&buf->st, buf->remote);
   if (i != data_end()) {
     stptr = data_node_get(i);
-    /* From chown(2): If  the owner or group is specified as -1, 
-       then that ID is not changed. 
+    /* From chown(2): If  the owner or group is specified as -1,
+       then that ID is not changed.
        Cannot put that test in libtricks, as at that point it isn't
        known what the fake user/group is (so cannot specify `unchanged')
-       
+
        I typecast to (uint32_t), as st.uid may be bigger than uid_t.
        In that case, the msb in st.uid should be discarded.
        I don't typecaset to (uid_t), as the size of uid_t may vary
@@ -625,7 +740,7 @@ void process_chown(struct fake_msg *buf){
        concurrently. Yes, this does seem farfeched, but was
        actually the case with the libc5/6 transition.
     */
-    if ((uint32_t)buf->st.uid != (uint32_t)-1) 
+    if ((uint32_t)buf->st.uid != (uint32_t)-1)
       stptr->uid=buf->st.uid;
     if ((uint32_t)buf->st.gid != (uint32_t)-1)
       stptr->gid=buf->st.gid;
@@ -646,11 +761,11 @@ void process_chown(struct fake_msg *buf){
 void process_chmod(struct fake_msg *buf){
   struct fakestat *st;
   data_node_t *i;
-  
+
   if(debug)
     fprintf(stderr,"FAKEROOT: chmod, mode=%lo\n",
 	    (long)buf->st.mode);
-  
+
   i = data_find(&buf->st, buf->remote);
   if (i != data_end()) {
     st = data_node_get(i);
@@ -674,7 +789,7 @@ void process_chmod(struct fake_msg *buf){
     if ((buf->st.mode&S_IFMT) != (st->mode&S_IFMT) &&
         ((buf->st.mode&S_IFMT) != S_IFREG || (!st->mode&(S_IFBLK|S_IFCHR)))) {
       fprintf(stderr,"FAKEROOT: chmod mode=%lo incompatible with "
-              "existing mode=%lo\n", buf->st.mode, st->mode);
+              "existing mode=%lo\n", (unsigned long)buf->st.mode, (unsigned long)st->mode);
       st->mode = buf->st.mode;
     }
     else{
@@ -692,11 +807,11 @@ void process_chmod(struct fake_msg *buf){
 void process_mknod(struct fake_msg *buf){
   struct fakestat *st;
   data_node_t *i;
-  
+
   if(debug)
     fprintf(stderr,"FAKEROOT: mknod, mode=%lo\n",
 	    (long)buf->st.mode);
-  
+
   i = data_find(&buf->st, buf->remote);
   if (i != data_end()) {
     st = data_node_get(i);
@@ -731,7 +846,7 @@ void process_stat(struct fake_msg *buf){
     cpyfakefake(&buf->st, data_node_get(i));
     if(debug){
       fprintf(stderr,"FAKEROOT: (previously known): fake=");
-      debug_stat(&buf->st);      
+      debug_stat(&buf->st);
     }
 
   }
@@ -760,6 +875,177 @@ void process_unlink(struct fake_msg *buf){
   }
 }
 
+void process_listxattr(struct fake_msg *buf)
+{
+#if defined(HAVE_LISTXATTR) || defined(HAVE_LLISTXATTR) || defined(HAVE_FLISTXATTR)
+  data_node_t *i;
+  xattr_node_t *x = NULL;
+
+  buf->xattr.flags_rc = 0;
+  i = data_find(&buf->st, buf->remote);
+  if(debug){
+    fprintf(stderr,"FAKEROOT: process listxattr\n");
+  }
+  if (i != data_end()) {
+    x = i->xattr;
+  }
+  if (!x) {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT:    (previously unknown)\n");
+    }
+    buf->xattr.buffersize = 0;
+  } else {
+    int bsize = 0;
+    while (x) {
+      int keysize = strlen(x->key);
+      if ((bsize + keysize + 1) > MAX_IPC_BUFFER_SIZE)
+      {
+        buf->xattr.flags_rc = ERANGE;
+        break;
+      }
+      strcpy(&buf->xattr.buf[bsize], x->key);
+      bsize += keysize + 1;
+      x = x->next;
+    }
+    buf->xattr.buffersize = bsize;
+    if(debug) {
+      fprintf(stderr,"FAKEROOT: (previously known): xattr=%s\n", buf->xattr.buf);
+    }
+  }
+  faked_send_fakem(buf);
+#endif /* defined(HAVE_LISTXATTR) || defined(HAVE_LLISTXATTR) || defined(HAVE_FLISTXATTR) */
+}
+
+void process_setxattr(struct fake_msg *buf)
+{
+#if defined(HAVE_SETXATTR) || defined(HAVE_LSETXATTR) || defined(HAVE_FSETXATTR)
+  data_node_t *i;
+  xattr_node_t *x = NULL;
+  xattr_node_t **x_ref = NULL;
+  xattr_node_t *new_node = NULL;
+  struct fakestat st;
+  char *value = NULL;
+  int key_size, value_size;
+  int flags = buf->xattr.flags_rc;
+
+  buf->xattr.flags_rc = 0;
+  /* Need some more bounds checking */
+  key_size = strlen(buf->xattr.buf);
+  value = &buf->xattr.buf[key_size + 1];
+  value_size = buf->xattr.buffersize - key_size - 1;
+
+  i = data_find(&buf->st, buf->remote);
+  if(debug){
+    fprintf(stderr,"FAKEROOT: process setxattr key = %s\n", buf->xattr.buf);
+  }
+  if (i == data_end()) {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT:    (previously unknown)\n");
+    }
+    st=buf->st;
+    /* We pretend that unknown files are owned
+       by root.root, so we have to maintain that pretense when the
+       caller asks to leave an id unchanged. */
+    if ((uint32_t)st.uid == (uint32_t)-1)
+       st.uid = 0;
+    if ((uint32_t)st.gid == (uint32_t)-1)
+       st.gid = 0;
+    insert_or_overwrite(&st, buf->remote);
+    i = data_find(&buf->st, buf->remote);
+  }
+  x = xattr_find(i->xattr, buf->xattr.buf);
+  if (x) {
+    if (flags == XATTR_CREATE) {
+      buf->xattr.flags_rc = EEXIST;
+      if (debug) {
+        fprintf(stderr,"FAKEROOT:    Already exists\n");
+      }
+    } else {
+      xattr_fill(x, buf->xattr.buf, value, value_size);
+      if (debug) {
+        fprintf(stderr,"FAKEROOT:    Replaced\n");
+      }
+    }
+  } else {
+    if (flags == XATTR_REPLACE) {
+      buf->xattr.flags_rc = ENODATA;
+      if (debug) {
+        fprintf(stderr,"FAKEROOT:    Replace requested but no previous entry found\n");
+      }
+    } else {
+      x = xattr_insert(&i->xattr);
+      xattr_fill(x, buf->xattr.buf, value, value_size);
+      if (debug) {
+        fprintf(stderr,"FAKEROOT:    Inserted\n");
+      }
+    }
+  }
+  buf->xattr.buffersize = 0;
+  faked_send_fakem(buf);
+#endif /* defined(HAVE_SETXATTR) || defined(HAVE_LSETXATTR) || defined(HAVE_FSETXATTR) */
+}
+
+void process_getxattr(struct fake_msg *buf)
+{
+#if defined(HAVE_GETXATTR) || defined(HAVE_LGETXATTR) || defined(HAVE_FGETXATTR)
+  data_node_t *i;
+  xattr_node_t *x = NULL;
+
+  buf->xattr.flags_rc = ENODATA;
+  i = data_find(&buf->st, buf->remote);
+  if(debug){
+    fprintf(stderr,"FAKEROOT: process getxattr key = %s\n", buf->xattr.buf);
+  }
+  if (i != data_end()) {
+    x = xattr_find(i->xattr, buf->xattr.buf);
+  }
+  if (!x) {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT:    (previously unknown)\n");
+    }
+    buf->xattr.buffersize = 0;
+  } else {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT: (previously known): %s\n", x->value);
+    }
+    buf->xattr.buffersize = x->value_size;
+    memcpy(buf->xattr.buf, x->value, x->value_size);
+    buf->xattr.flags_rc = 0;
+  }
+  faked_send_fakem(buf);
+#endif /* defined(HAVE_GETXATTR) || defined(HAVE_LGETXATTR) || defined(HAVE_FGETXATTR) */
+}
+
+void process_removexattr(struct fake_msg *buf)
+{
+#if defined(HAVE_REMOVEXATTR) || defined(HAVE_LREMOVEXATTR) || defined(HAVE_FREMOVEXATTR)
+  data_node_t *i;
+  xattr_node_t *x = NULL;
+
+  buf->xattr.flags_rc = ENODATA;
+  i = data_find(&buf->st, buf->remote);
+  if(debug){
+    fprintf(stderr,"FAKEROOT: process removexattr key = %s\n", buf->xattr.buf);
+  }
+  if (i != data_end()) {
+    x = xattr_find(i->xattr, buf->xattr.buf);
+  }
+  if (!x) {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT:    (previously unknown)\n");
+    }
+  } else {
+    if (debug) {
+      fprintf(stderr,"FAKEROOT: (previously known): %s\n", x->value);
+    }
+    xattr_erase(&i->xattr, buf->xattr.buf);
+    buf->xattr.flags_rc = 0;
+  }
+  buf->xattr.buffersize = 0;
+  faked_send_fakem(buf);
+#endif /* defined(HAVE_REMOVEXATTR) || defined(HAVE_LREMOVEXATTR) || defined(HAVE_FREMOVEXATTR) */
+}
+
 void debugdata(int dummy UNUSED){
   int stored_errno = errno;
   data_node_t *i;
@@ -793,13 +1079,14 @@ void get_msg()
     r=msgrcv(msg_get,&buf,sizeof(struct fake_msg),0,0);
     if(debug)
       fprintf(stderr,"FAKEROOT: r=%i, received message type=%li, message=%i\n",r,buf.mtype,buf.id);
-    if(r!=-1)
+    if(r!=-1) {
       buf.remote = 0;
       process_msg(&buf);
+    }
   }while ((r!=-1)||(errno==EINTR));
   if(debug){
     perror("FAKEROOT, get_msg");
-    fprintf(stderr,"r=%i, EINTR=%i\n",errno,EINTR);
+    fprintf(stderr,"errno=%i, EINTR=%i\n",errno,EINTR);
   }
 }
 
@@ -1003,6 +1290,8 @@ static int get_fakem(struct fake_msg *buf)
   buf->st.mode = ntohl(buf->st.mode);
   buf->st.nlink = ntohl(buf->st.nlink);
   buf->remote = ntohl(buf->remote);
+  buf->xattr.buffersize = ntohl(buf->xattr.buffersize);
+  buf->xattr.flags_rc = ntohl(buf->xattr.flags_rc);
 
   return 0;
 }
@@ -1017,6 +1306,7 @@ int main(int argc, char **argv){
 #ifndef FAKEROOT_FAKENET
   union semun sem_union;
   int justcleanup = 0;
+  int msgflag = IPC_CREAT|0600;
 #else /* FAKEROOT_FAKENET */
   int sd, val;
   unsigned int port = 0;
@@ -1087,29 +1377,43 @@ int main(int argc, char **argv){
 
 #ifndef FAKEROOT_FAKENET
 
-  if(!msg_key) {
-    srandom(time(NULL)+getpid()*33151);
-    while(!msg_key && (msg_key!=-1))  /* values 0 and -1 are treated
-					 specially by libfake */
-      msg_key=random();
-  }
-    
-  if(debug)
-    fprintf(stderr,"using %li as msg key\n",(long)msg_key);
-  
-  msg_get=msgget(msg_key,IPC_CREAT|0600);
-  msg_snd=msgget(msg_key+1,IPC_CREAT|0600);
-  sem_id=semget(msg_key+2,1,IPC_CREAT|0600);
+  do {
+    if(!msg_key) {
+      msgflag |= IPC_EXCL;
+      srandom(time(NULL)+getpid()*33151);
+      while(!msg_key && (msg_key!=-1))  /* values 0 and -1 are treated
+                                           specially by libfake */
+        msg_key=random();
+    }
+
+    if(debug)
+      fprintf(stderr,"using %li as msg key\n",(long)msg_key);
+
+    msg_get=msgget(msg_key,msgflag);
+    if (msg_get != -1)
+      msg_snd=msgget(msg_key+1,msgflag);
+    if (msg_snd != -1)
+      sem_id=semget(msg_key+2,1,msgflag);
+
+    if((msg_get==-1)||(msg_snd==-1)||(sem_id==-1)){
+      if (errno == EEXIST) {
+        if(debug)
+          fprintf(stderr,"using %li as msg key caused a collision, cleaning up and retrying\n",(long)msg_key);
+        cleanup(-1);
+        msg_key = 0;
+        msg_get = msg_snd = sem_id = -1;
+        continue;
+      }
+      perror("fakeroot, while creating message channels");
+      fprintf(stderr, "This may be due to a lack of SYSV IPC support.\n");
+      cleanup(-1);
+      exit(1);
+    }
+  } while(msg_key == 0);
+
   sem_union.val=1;
   semctl (sem_id,0,SETVAL,sem_union);
 
-  if((msg_get==-1)||(msg_snd==-1)||(sem_id==-1)){
-    perror("fakeroot, while creating message channels");
-    fprintf(stderr, "This may be due to a lack of SYSV IPC support.\n");
-    cleanup(-1);
-    exit(1);
-  }
-
   if(debug)
     fprintf(stderr,"msg_key=%li\n",(long)msg_key);
 
@@ -1133,7 +1437,7 @@ int main(int argc, char **argv){
   if (port > 0) {
     memset((char *) &addr, 0, sizeof (addr));
     addr.sin_family = AF_INET;
-    addr.sin_addr.s_addr = htonl(INADDR_ANY);
+    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
     addr.sin_port = htons(port);
 
     if (bind(sd, (struct sockaddr *) &addr, sizeof (addr)) < 0)
@@ -1164,7 +1468,7 @@ int main(int argc, char **argv){
   sigemptyset(&sa_debug.sa_mask);
   sa_debug.sa_flags=0;
   //  sa_debug.sa_restorer=0;
-  
+
   sa_save.sa_handler=save;
   sigemptyset(&sa_save.sa_mask);
   sa_save.sa_flags=0;
@@ -1199,7 +1503,7 @@ int main(int argc, char **argv){
     if ((pid=fork()) == 0){
       int fl;
       int num_fds = getdtablesize();
-      
+
       fflush(stdout);
 
       /* This is the child closing its file descriptors. */
